"""Simple Slack Store using Webhooks
For setup see: https://api.slack.com/messaging/webhooks
"""

from sinai.stores.api import ApiStore
from sinai.stores.base import MetricStore
from sinai.types import MetricInstance


class Slack(ApiStore, MetricStore):
    # An example, set your own.
    url = (
        "https://hooks.slack.com/services/T00000000/B00000000/XXXXXXXXXXXXXXXXXXXXXXXX"
    )

    def save_metric(self, metric: MetricInstance) -> None:
        self.post({"message": metric.format_message()})



"""
Publishes sinai metrics as metric data points in Amazon CloudWatch.

For more information:

https://boto3.amazonaws.com/v1/documentation/api/latest/reference/
services/cloudwatch.html#CloudWatch.Client.put_metric_data
"""

try:
    import boto3
except ImportError:
    HAS_BOTO = False
else:
    HAS_BOTO = True

from sinai.exceptions import DependencyMissing
from sinai.stores.base import Store
from sinai.types import MetricInstance, MonitorInstance


class CloudWatch(Store):
    namespace = "metrics"
    dimension = "context"

    def __init__(self, monitor: MonitorInstance):
        super().__init__(monitor)
        if not HAS_BOTO:
            raise DependencyMissing(
                "Sorry, boto3 library required to store metrics in Cloudwatch."
            )

    def save_metric(self, metric: MetricInstance) -> None:
        client = boto3.client("cloudwatch")
        client.put_metric_data(
            Namespace=self.namespace,
            MetricData=[
                {
                    "MetricName": metric.name,
                    "Value": metric.value,
                    "Unit": "Count",
                    "Dimensions": [{"Name": self.dimension, "Value": metric.context}],
                }
            ],
        )
        return super().save_metric(metric)



from sinai.stores.memory import MemoryMetricStore


class DebugStore(MemoryMetricStore):
    """Prints metrics out to standard output."""

    def save_metric(self, metric):
        print("DEBUG: We have a metric!\n")
        print(metric.format_message())
        super().save_metric(metric)



"""Custom Data Types for Sinai."""
from __future__ import annotations

from decimal import Decimal
from numbers import Number
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Sequence,
    Tuple,
    Type,
    Union,
)
from uuid import UUID

try:
    # Version 3.10 and above:
    from typing import TypeAlias
except ImportError:
    # Python 3.9 and below:
    try:
        from typing_extensions import TypeAlias
    except ImportError:
        # Don't need typing in runtime
        TypeAlias = None  # type: ignore


if TYPE_CHECKING:  # pragma: no cover
    from sinai.metrics import AggregationMetric, Metric  # noqa: F401
    from sinai.monitors import Monitor  # noqa: F401
    from sinai.rules import Rule
    from sinai.sources import MetricSource, Source  # noqa: F401
    from sinai.stores import Store

    try:
        from bson.objectid import ObjectId  # pylint: disable=W0611
    except ImportError:
        pass

    try:
        from requests import Response  # noqa: F401
    except ImportError:
        pass


# Common Data Types
JDict = Dict[str, Any]  # A JSON-Safe Dictionary
JDictOrNone = Union[JDict, None]  # A JSON-Safe Dictionary or none
FList = List[Any]  # A Freeform List
SList = List[str]  # A List of Strings
NullableBool = Union[bool, None]

# Metric types
MetricInstance: TypeAlias = "Metric"
MetricList = List[MetricInstance]
MetricClass = Type[MetricInstance]
MetricClasses = List[MetricClass]
MetricId = Union[UUID, "ObjectId", str]
MetricDict = Dict[MetricId, MetricInstance]
MetricResult = Tuple[MetricInstance, MetricId]
MetricValue = Union[Number, Decimal, None, int, float]
MetricValueList = List[MetricValue]
AggregationMetricInstance: TypeAlias = "AggregationMetric"
AggregationMetricClass = Type[AggregationMetricInstance]
AggregationFunction = Callable[[MetricValueList], MetricValue]
AggregationMetrics = List["AggregationMetric"]
Evaluation = Union[Sequence[MetricInstance], AggregationMetrics, MetricInstance, None]

# Monitor Types
MonitorInstance: TypeAlias = "Monitor"
GlobalMemory = Dict[MetricId, JDict]

# Rule types
RuleClass = Type["Rule"]
RuleClasses = List[RuleClass]
RuleInstance: TypeAlias = "Rule"
RuleInstances: Sequence[RuleInstance]

# Store types
StoreClass = Type["Store"]
StoreClassList = List[StoreClass]
StoreDict = Dict[StoreClass, "Store"]
StoreClasses = Sequence[StoreClass]

# Source types
SourceInstance: TypeAlias = "Source"
SourceInstances = Sequence[SourceInstance]
SourceClass = Type[SourceInstance]
SourceClassList = List[SourceClass]
SourceDict = Dict[SourceClass, SourceInstance]
SourceClasses = Sequence[SourceClass]
OptionalSourceInstances = Union[Sequence[SourceInstance], None]

# MetricSource types
MetricSourceInstance: TypeAlias = "MetricSource"
MetricSourceClass = Type[MetricSourceInstance]
MetricSourceClasses = Sequence[MetricSourceClass]

# ApiSource types
RequestHeader = Dict[str, str]

# Mongo adapter types
MongoResult: TypeAlias = Any

# ApiStore types
RequestResponse: TypeAlias = "Response"
PostResponse = Union[JDict, bytes]



"""
The Monitor connects to sources, evaluates rules, sends the resulting metrics to the stores.
"""
from __future__ import annotations

import sys
from collections.abc import Iterable
from typing import get_type_hints
from uuid import uuid4

from sinai.exceptions import SourceError, SourceNotFound
from sinai.metrics import Metric
from sinai.rules import Rule
from sinai.sources import MetricSource
from sinai.types import (
    Evaluation,
    JDict,
    OptionalSourceInstances,
    RuleClass,
    RuleClasses,
    SourceClass,
    SourceClassList,
    SourceDict,
    SourceInstance,
    StoreClass,
    StoreClassList,
    StoreDict,
)


class Monitor:
    """The monitor controls the monitoring run."""

    rules: RuleClasses = []
    context: str = "global"

    def __init__(self) -> None:
        self.id = str(uuid4())
        self.sources: SourceClassList = []
        self.stores: StoreClassList = []
        self._store_instances: StoreDict = {}
        self._source_instances: SourceDict = {}
        self._source_map = {}  # type: ignore
        self._resolve_rules()

    def _add_source(self, source_cls: SourceClass) -> None:
        if source_cls not in self.sources:
            self.sources.append(source_cls)

    def _add_store(self, store_cls: StoreClass) -> None:
        if store_cls not in self.stores:
            self.stores.append(store_cls)

    def __resolve_stores_from_attribute(self, rule_cls: RuleClass) -> None:
        """Get the stores from stores = []"""
        for store_cls in rule_cls.stores:
            self._add_store(store_cls)

    def __resolve_sources_from_attribute(self, rule_cls: RuleClass) -> None:
        """Get the sources from sources = []"""
        for source_cls in rule_cls.sources:
            self._add_source(source_cls)

    def _get_annotations(self, rule_cls: RuleClass) -> JDict:
        """Getting annonations for the constructor."""
        if sys.version_info.minor > 8:
            try:
                return get_type_hints(rule_cls.__init__, include_extras=True)
            except NameError:
                pass

        try:
            return get_type_hints(rule_cls.__init__)
        except NameError:
            pass

        return rule_cls.__init__.__annotations__

    def __resolve_sources_from_init(self, rule_cls: RuleClass) -> None:
        self._source_map[rule_cls] = {}
        annos = self._get_annotations(rule_cls)
        for arg, source_cls in annos.items():
            if arg == "monitor":
                self._source_map[rule_cls]["monitor"] = self.__class__
            elif arg == "return":
                continue
            elif source_cls == "MetricSource":
                self._add_source(MetricSource)
                self._source_map[rule_cls][arg] = MetricSource
            else:
                self._add_source(source_cls)
                self._source_map[rule_cls][arg] = source_cls

    def _resolve_rules(self) -> None:
        for rule_cls in self.rules:
            self.__resolve_stores_from_attribute(rule_cls)
            self.__resolve_sources_from_attribute(rule_cls)
            self.__resolve_sources_from_init(rule_cls)

    def _connect_sources(self) -> None:
        for source_class in self.sources:
            try:
                self._source_instances[source_class] = source_class(monitor=self)
            except TypeError:
                raise SourceError(
                    f"The object {source_class} is not a valid Source class."
                )

    def _connect_stores(self) -> None:
        for store_class in self.stores:
            self._store_instances[store_class] = store_class(monitor=self)

    def _evaluate_rules(self) -> None:
        for rule_class in self.rules:
            self._evaluate_rule(rule_class)

    def _get_rule_sources(self, rule_cls: RuleClass) -> OptionalSourceInstances:
        return [self.source(source) for source in rule_cls.sources]

    def _get_rule_args(self, rule_cls: RuleClass):
        kwargs: JDict = {}
        for arg, source_cls in self._source_map[rule_cls].items():
            if arg == "monitor":
                kwargs["monitor"] = self
            else:
                kwargs[arg] = self.source(source_cls)

        return kwargs

    def _evaluate_rule(self, rule_class: RuleClass) -> None:
        kwargs = self._get_rule_args(rule_class)
        rule = rule_class(**kwargs)
        if "monitor" not in kwargs:
            rule.monitor = self
        result: Evaluation = rule.evaluate()
        if isinstance(result, Iterable):
            for metric in result:
                self.store_metric(rule, metric)
        elif not result:
            return
        else:
            self.store_metric(rule, result)

    def execute(self) -> None:
        """Start monitoring."""
        self._connect_stores()
        self._connect_sources()
        self._evaluate_rules()

    def source(self, source: SourceClass) -> SourceInstance:
        """Return an instatatied source."""
        try:
            return self._source_instances[source]
        except KeyError:
            raise SourceNotFound(f"Could not find {source.__name__} source instance.")

    def store_metric(self, rule: Rule, metric: Metric) -> None:
        """Store a metric."""
        if not metric.context:
            metric.context = self.context
        for store_class in rule.stores:
            store = self._store_instances[store_class]
            store.save_metric(metric)



"""Support for using Mongo Databases as a source or store for data."""

from __future__ import annotations

try:
    import pymongo
except ImportError:
    HAS_MONGO = False
else:
    HAS_MONGO = True

try:
    import certifi
except ImportError:
    HAS_CERTIFI = False
else:
    HAS_CERTIFI = True

from sinai.exceptions import DependencyMissing
from sinai.types import MonitorInstance


class Mongo:
    """A connection to MongoDB."""

    connection_string = ""
    database_name = ""
    use_certificates = False

    def __init__(self, monitor: MonitorInstance) -> None:
        self.monitor = monitor
        if not HAS_MONGO:
            raise DependencyMissing(
                "Pymongo library required for using Mongo databases."
            )
        if self.use_certificates:
            if not HAS_CERTIFI:
                raise DependencyMissing(
                    "The certifi library is required for using SSL certifications with Mongo database connections."
                )
            self.client = pymongo.MongoClient(self.connection_string, tlsCAFile=certifi.where())  # type: ignore
        else:
            self.client = pymongo.MongoClient(self.connection_string)  # type: ignore
        self.db = self.client[self.database_name]
        self.collection = self.db["metrics"]



"""
Support for using PostgreSQL as a source or store for data.
"""

try:
    import psycopg2
except ImportError:
    HAS_POSTGRES = False
else:
    HAS_POSTGRES = True


from sinai.exceptions import DependencyMissing
from sinai.types import MonitorInstance


class Postgresql:
    """Connects to from PostgreSQL databases."""

    connection_string = "dbname=test user=postgres"

    def __init__(self, monitor: MonitorInstance):
        self.monitor = monitor
        if not HAS_POSTGRES:
            raise DependencyMissing(
                "psycopg2 library required for using PostgreSQL databases."
            )
        self.conn = psycopg2.connect(self.connection_string)

    def __del__(self):
        self.conn.close()



__all__ = [
    "Store",
    "MetricStore",
    "MongoMetricStore",
    "MemoryMetricStore",
    "ApiStore",
    "CloudWatch",
    "Slack",
    "DebugStore",
]
from sinai.stores.api import ApiStore
from sinai.stores.base import MetricStore, Store
from sinai.stores.cloudwatch import CloudWatch
from sinai.stores.debug import DebugStore
from sinai.stores.memory import MemoryMetricStore
from sinai.stores.mongo import MongoMetricStore
from sinai.stores.slack import Slack



__all__ = [
    "Source",
    "MetricSource",
    "MongoSource",
    "MongoMetricSource",
    "ApiSource",
    "PostgresSource",
]

from sinai.sources.api import ApiSource
from sinai.sources.base import MetricSource, Source
from sinai.sources.mongo import MongoMetricSource, MongoSource
from sinai.sources.postgresql import PostgresSource



"""A Rule takes data from sources, makes metrics and puts them into stores."""
from __future__ import annotations

from sinai.types import (
    Evaluation,
    MonitorInstance,
    SourceClasses,
    SourceInstance,
    StoreClasses,
)


class Rule:
    """The sources and stores are defined in the rule, instantiated by the monitor."""

    sources: SourceClasses = []
    stores: StoreClasses = []

    def __init__(self, monitor: MonitorInstance, *sources: SourceInstance):
        self.monitor = monitor

    def evaluate(self) -> Evaluation:
        """The evalutation function is called by the monitor, which stores any returned metrics."""
        return None



"""The Api store posts data to an API endpoint.
Requires the requests library to be installed.
"""

from sinai.adaptors.api import ApiConnection
from sinai.exceptions import SourceError
from sinai.stores.base import Store
from sinai.types import JDict, MonitorInstance, PostResponse, RequestResponse


class ApiStore(Store, ApiConnection):
    """Send data to an API endpoint."""

    def __init__(self, monitor: MonitorInstance):
        super().__init__(monitor)
        self.content: JDict = {}
        self.response: RequestResponse = self.requests.Response()

    def post(self, data: JDict) -> PostResponse:
        """Post the data to the specified API."""
        self.response = self.requests.post(
            self.url, headers=self.get_headers(), json=data
        )
        if self.response.status_code != 200:
            raise SourceError(
                f"The URL {self.url} returned status {self.response.status_code}."
            )
        try:
            json_response: JDict = self.response.json()
        except self.requests.JSONDecodeError:
            return self.response.content
        return json_response



"""
Support for using PostgreSQL as a source of data.
"""

from sinai.adaptors.postgresql import Postgresql
from sinai.sources.base import Source


class PostgresSource(Postgresql, Source):
    """Gets data from PostgreSQL databases."""

    ...



"""A Metric is the output of a rule, and often the input to rules too.
Metrics can be stored in a Store, and can be retrieved by a MetricSource."""
import datetime
from uuid import uuid4

from sinai.types import JDict, MetricValue, SList


class Metric:
    """A measure to be observed."""

    name: str = "monitoring_metric"
    context: str = ""
    update: SList = []

    def __init__(
        self,
        value: MetricValue = None,
        name: str = "",
        ref: str = "",
        update: SList = [],
        checked: bool = False,
    ) -> None:
        if name:
            self.name = name
        if update:
            self.update = update
        self.annotations: SList = []
        self.monitor_id: str = ""
        self.ref = ref or str(uuid4())
        self.value = value
        self.checked = checked
        self.created_at = datetime.datetime.utcnow()
        self.updated_at = self.created_at

    def pre_save(self, monitor_id: str) -> JDict:
        """Readythe Metric for storage."""
        self.updated_at = datetime.datetime.utcnow()
        self.monitor_id = monitor_id
        return self.to_dict()

    def to_dict(self) -> JDict:
        """Serialise the Metric to a dictionary."""
        return {
            "name": self.name,
            "ref": self.ref,
            "value": self.value,
            "context": self.context,
            "annotations": self.annotations,
            "created_at": self.created_at,
            "updated_at": self.updated_at,
            "monitor_id": self.monitor_id,
            "checked": self.checked,
        }

    @classmethod
    def from_dict(cls, metric_dict: JDict) -> "Metric":
        """Deserialise a dictionary back to a Metric."""
        metric = cls(
            name=metric_dict["name"],
            ref=metric_dict["ref"],
            value=metric_dict["value"],
            checked=metric_dict["checked"],
        )
        metric.context = metric_dict["context"]
        metric.created_at = metric_dict["created_at"]
        metric.updated_at = metric_dict["updated_at"]
        metric.annotations = metric_dict["annotations"]
        metric.monitor_id = metric_dict["monitor_id"]
        return metric

    def annotate(self, text: str) -> None:
        """Store a piece of text in the Metric."""
        self.annotations.append(text)

    def format_message(self):
        """Format the metric into text."""
        message = ""
        if self.context:
            message += f"{self.context}: "
        message += f"{self.name}\n"
        message += f"{self.__doc__}\n"
        if self.ref:
            message += f"Ref: {self.ref}\n"
        if self.value:
            message += f"Value: {self.value}\n"
        message += f"First Seen: {self._prettify_created_at()}\n"
        if self.checked:
            message += "Checked\n"
        if self.annotations:
            message += "Notes:\n"
            for note in self.annotations:
                message += note + "\n"
        return message

    def _prettify_created_at(self) -> str:
        formatted_date = self.created_at.strftime("%Y-%m-%d")
        if self.created_at.date() == datetime.date.today():
            return f"{formatted_date} - NEW TODAY!"
        else:
            return f"{formatted_date} - {(self.created_at - datetime.datetime.utcnow()).days} days ago"


class Sinner(Metric):
    """An unexpected or missing item between two sequences."""

    ...
